'use strict';

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _os = require('os');

var _indentString = require('indent-string');

var _indentString2 = _interopRequireDefault(_indentString);

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _funcHasParam = require('func-has-param');

var _funcHasParam2 = _interopRequireDefault(_funcHasParam);

var _path = require('path');

var _newlineRegex = require('newline-regex');

var _newlineRegex2 = _interopRequireDefault(_newlineRegex);

var _countSpaces = require('count-spaces');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Returns whether the module's config function is injecting parameter
 * @param {String} fileContents - file contents of module
 * @param {Object} config - user's config
 * @return {Boolean} - does module have param?
 */
var hasParam = function hasParam(fileContents, config) {
  var param = config.ngRoute ? '$routeProvider' : '$stateProvider';

  var opts = config.appScript === 'ts' ? { language: 'ts' } : null;

  if (config.appScript === 'coffee') {
    return new RegExp('\\(.*' + param + '.*\\) ->').test(fileContents);
  }

  return (0, _funcHasParam2.default)(fileContents, 'config', param, opts);
};

/**
 * Line includes config function
 * @param {String} line - line to check for config function existence
 * @param {Boolean} isCoffeeScript - is the line written in CoffeeScript?
 * @return {Boolean} - line has config function
 */
var lineHasConfigFunction = function lineHasConfigFunction(line, isCoffeeScript) {
  if (isCoffeeScript) {
    return line.indexOf('config') > -1 && line.indexOf('->') > -1;
  }

  return line.indexOf('function config(') > -1;
};

/**
 * Add param to config function line
 * @param {String} line - line to add parameter to
 * @param {String} param - parameter to add
 * @param {String} lang - which language is line written in? (coffee, es6, js, or ts)
 * @param {String} [type] - when lang === ts then insert type
 * @return {String} - config function line with parameter added
 */
var addParamToConfigFunction = function addParamToConfigFunction(line, param, lang, type) {
  var paramToInsert = '$' + param;

  // config line ending will be stripped
  // this will add it back
  if (lang === 'ts') {
    paramToInsert += ': ' + type + ') {';
  } else if (lang === 'coffee') {
    paramToInsert += ') ->';
  } else {
    paramToInsert += ') {';
  }

  // previous state had no parameters - config()
  if (line.lastIndexOf('(') === line.lastIndexOf(')') - 1) {
    return line.slice(0, line.lastIndexOf(')')) + paramToInsert;
  }

  // CoffeScript && previous state had no existing parameters - config ->
  if (lang === 'coffee' && line.lastIndexOf('(') === -1 && line.lastIndexOf(')') === -1) {
    return line.slice(0, line.lastIndexOf('g')) + 'g (' + paramToInsert;
  }

  // previous state had existing parameters - config()
  return line.slice(0, line.lastIndexOf(')')) + ', ' + paramToInsert;
};

/**
 * Injects the parameter into the module's config function
 * @param {Array} lines - lines to modify
 * @param {Object} config - user's config
 * @return {Array} - modified lines
 */
var addParam = function addParam(lines, config) {
  var param = config.ngRoute ? 'routeProvider' : 'stateProvider',
      type = config.ngRoute ? 'ng.route.IRouteProvider' : 'ng.ui.IStateProvider';

  return lines.map(function (line) {
    if (lineHasConfigFunction(line, config.appScript === 'coffee')) {
      return addParamToConfigFunction(line, param, config.appScript, type);
    }

    return line;
  });
};

/**
 * Analyzes lines to determine where to insert new route
 * @param {Array} lines - lines to analyze
 * @param {Object} config - user's config
 * @return {Object} - info regarding insertion points
 */
var analyzeLines = function analyzeLines(lines, config) {
  var newRoute = config.ngRoute ? 'when' : 'state';

  var braceCount = 0;

  var analysis = {
    routeStartIndex: -1,
    routeEndIndex: -1,
    configFunctionIndex: -1
  };

  lines.forEach(function (line, i) {
    if (config.appScript === 'ts' || config.appScript === 'js' || config.appScript === 'es6') {
      if (line.indexOf('function config(') > -1) {
        analysis.configFunctionIndex = i;
      }

      // look for .state and set routeStartIndex
      if (line.indexOf('.' + newRoute + '(') > -1) {
        analysis.routeStartIndex = i;
      }

      // open braces add to braceCount
      if (analysis.routeStartIndex > -1 && line.indexOf('{') > -1) {
        braceCount++;
      }

      // close braces subract from braceCount
      if (analysis.routeStartIndex > -1 && line.indexOf('}') > -1) {
        braceCount--;
      }

      // when braceCount = 0 the end of the state has been reached
      // set routeEndIndex
      if (analysis.routeStartIndex > -1 && braceCount === 0) {
        analysis.routeEndIndex = i;
      }
    } else {
      if (line.indexOf('.config') > -1 && line.indexOf('->') > -1) {
        analysis.configFunctionIndex = i;
      }

      // look for .state and set routeStartIndex
      if (line.indexOf('.' + newRoute) > -1) {
        analysis.routeStartIndex = i;
      }
    }
  });

  analysis.existingRouteFound = analysis.routeStartIndex !== -1;

  return analysis;
};

/**
 * Returns new state to add
 * @param {Object} state - state info to format
 * @param {Object} analysis - information for insertion
 * @param {Object} config - user's config
 * @return {String[]} - lines of new state
 */
var prepareState = function prepareState(state, analysis, config) {
  var templateFile = void 0;

  var context = {
    analysis: analysis,
    config: config,
    state: state
  };

  templateFile = (0, _path.join)(__dirname, 'templates', '_' + (config.ngRoute ? 'ngroute' : 'uirouter') + '.');
  // js, es6, ts files all use the same templates
  templateFile += config.appScript === 'coffee' ? 'coffee' : 'js';

  return _lodash2.default.template(_fs2.default.readFileSync(templateFile))(context).split(_newlineRegex2.default);
};

/**
 * Adds state to lines
 * @param {Array} lines - lines to modify with state
 * @param {Object} state - state info to add to lines
 * @param {Object} analysis - insertion info
 * @param {Object} config - user's config
 * @return {Array} - modified lines with added state
 */
var addState = function addState(lines, state, analysis, config) {
  var insertLine = void 0,
      numOfSpaces = void 0,
      numOfSpacesCounter = void 0;

  // count spaces to prepend to state
  var lineToCheck = lines[analysis.existingRouteFound ? analysis.routeStartIndex : analysis.configFunctionIndex];

  numOfSpaces = (0, _countSpaces.start)(lineToCheck);

  // if there is an existing route, remove 2 more spaces from indent to line up with existing routes
  // else add 2 spaces
  if (analysis.existingRouteFound) {
    numOfSpaces -= 2;
  } else {
    numOfSpaces += 2;
  }

  // prepend spaces
  state = state.map(function (stateLine) {
    return (0, _indentString2.default)(stateLine, ' ', numOfSpaces);
  });

  if (config.appScript !== 'coffee') {
    insertLine = analysis.existingRouteFound ? analysis.routeEndIndex : analysis.configFunctionIndex + 1;
  } else if (analysis.existingRouteFound) {
    // determine where last state ends by examining spaces
    // insert new route on first line to have less spaces at the start
    numOfSpaces = (0, _countSpaces.start)(lines[analysis.routeStartIndex]);
    numOfSpacesCounter = numOfSpaces;
    insertLine = analysis.routeStartIndex;
    while (numOfSpacesCounter >= numOfSpaces) {
      insertLine++;
      numOfSpacesCounter = (0, _countSpaces.start)(lines[insertLine]);
    }
  } else {
    insertLine = analysis.configFunctionIndex + 1;
  }

  lines.splice(insertLine, 0, state.join(_os.EOL));

  return lines.join(_os.EOL);
};

module.exports = {
  /**
   * Adds route to module's config
   * @param {String} fileContents - file contents of module
   * @param {Object} state - state info to add
   * @param {Object} config - user's config
   * @return {String} - modified file contents with added state
   */

  addRoute: function addRoute(fileContents, state, config) {
    // checking if provider is used
    var needsParam = !hasParam(fileContents, config);

    var lines = fileContents.split(_newlineRegex2.default);

    var analysis = analyzeLines(lines, config);

    // if file needs parameter, add it
    if (needsParam) {
      lines = addParam(lines, config);
    }

    var newState = prepareState(state, analysis, config);

    return addState(lines, newState, analysis, config);
  }
};